@@ -41,9 +41,9 @@ Q_LOGGING_CATEGORY(lcEffectSystem, "qt.quick3d.effects");
 
 struct QSSGRhiEffectTexture
 {
-    QRhiTexture *texture;
-    QRhiRenderPassDescriptor *renderPassDescriptor;
-    QRhiTextureRenderTarget *renderTarget;
+    QRhiTexture *texture = nullptr;
+    QRhiRenderPassDescriptor *renderPassDescriptor = nullptr;
+    QRhiTextureRenderTarget *renderTarget = nullptr;
     QByteArray name;
 
     QSSGRhiSamplerDescription desc;
@@ -68,14 +68,13 @@ QSSGRhiEffectSystem::~QSSGRhiEffectSystem()
     releaseResources();
 }
 
-void QSSGRhiEffectSystem::setup(QRhi *, QSize outputSize, QSSGRenderEffect *firstEffect)
+void QSSGRhiEffectSystem::setup(QSize outputSize)
 {
-    if (!firstEffect || outputSize.isEmpty()) {
+    if (outputSize.isEmpty()) {
         releaseResources();
         return;
     }
     m_outSize = outputSize;
-    m_firstEffect = firstEffect;
 }
 
 QSSGRhiEffectTexture *QSSGRhiEffectSystem::findTexture(const QByteArray &bufferName)
@@ -140,7 +139,6 @@ QSSGRhiEffectTexture *QSSGRhiEffectSystem::getTexture(const QByteArray &bufferNa
         m_pendingClears.insert(result->renderTarget);
     } else if (needsRebuild) {
         if (formatChanged) {
-            m_rhiContext->invalidateCachedReferences(result->renderPassDescriptor);
             delete result->renderPassDescriptor;
             result->renderPassDescriptor = result->renderTarget->newCompatibleRenderPassDescriptor();
             result->renderTarget->setRenderPassDescriptor(result->renderPassDescriptor);
@@ -169,6 +167,7 @@ void QSSGRhiEffectSystem::releaseTextures()
 
 QRhiTexture *QSSGRhiEffectSystem::process(const QSSGRef<QSSGRhiContext> &rhiCtx,
                                           const QSSGRef<QSSGRenderer> &renderer,
+                                          const QSSGRenderEffect &firstEffect,
                                           QRhiTexture *inTexture,
                                           QRhiTexture *inDepthTexture,
                                           QVector2D cameraClipRange)
@@ -179,10 +178,9 @@ QRhiTexture *QSSGRhiEffectSystem::process(const QSSGRef<QSSGRhiContext> &rhiCtx,
         return inTexture;
     m_depthTexture = inDepthTexture;
     m_cameraClipRange = cameraClipRange;
-    Q_ASSERT(m_firstEffect);
 
     m_currentUbufIndex = 0;
-    auto *currentEffect = m_firstEffect;
+    auto *currentEffect = &firstEffect;
     QSSGRhiEffectTexture firstTex{ inTexture, nullptr, nullptr, {}, {}, {} };
     auto *latestOutput = doRenderEffect(currentEffect, &firstTex);
     firstTex.texture = nullptr; // make sure we don't delete inTexture when we go out of scope
@@ -194,13 +192,11 @@ QRhiTexture *QSSGRhiEffectSystem::process(const QSSGRef<QSSGRhiContext> &rhiCtx,
     }
 
     releaseTextures();
-    return latestOutput->texture;
+    return latestOutput ? latestOutput->texture : nullptr;
 }
 
 void QSSGRhiEffectSystem::releaseResources()
 {
-    for (auto *t : qAsConst(m_textures))
-        m_rhiContext->invalidateCachedReferences(t->renderPassDescriptor);
     qDeleteAll(m_textures);
     m_textures.clear();
     m_currentOutput = nullptr;
@@ -350,20 +346,18 @@ void QSSGRhiEffectSystem::applyInstanceValueCmd(const QSSGApplyInstanceValue *in
                         mipMode = QSSGBufferManager::MipModeGenerated;
                     // ### would we want MipModeBsdf in some cases?
 
-                    QSSGRenderImageTextureData theTextureData = theBufferManager->loadRenderImage(image, false, mipMode);
-
-                    if (theTextureData.m_rhiTexture) {
+                    const QSSGRenderImageTexture texture = theBufferManager->loadRenderImage(image, mipMode);
+                    if (texture.m_texture) {
                         const QSSGRhiSamplerDescription desc{
                             toRhi(textureProperty.minFilterType),
                             toRhi(textureProperty.magFilterType),
                             textureProperty.mipFilterType != QSSGRenderTextureFilterOp::None ? toRhi(textureProperty.mipFilterType) : QRhiSampler::None,
-                            toRhi(textureProperty.clampType),
-                            toRhi(textureProperty.clampType)
+                            toRhi(textureProperty.horizontalClampType),
+                            toRhi(textureProperty.verticalClampType)
                         };
-                        addTextureToShaderPipeline(textureProperty.name, theTextureData.m_rhiTexture, desc);
+                        addTextureToShaderPipeline(textureProperty.name, texture.m_texture, desc);
                         texAdded = true;
                     }
-                    image->m_textureData = theTextureData;
                 }
             }
             if (!texAdded) {
@@ -403,58 +397,75 @@ static const char *effect_builtin_textureMapUVFlipped =
         "    return vec2(uv.x, 1.0 - uv.y);\n"
         "}\n";
 
+QSSGRef<QSSGRhiShaderPipeline> QSSGRhiEffectSystem::buildShaderForEffect(const QSSGBindShader &inCmd,
+                                                                         const QSSGRef<QSSGProgramGenerator> &generator,
+                                                                         const QSSGRef<QSSGShaderLibraryManager> &shaderLib,
+                                                                         const QSSGRef<QSSGShaderCache> &shaderCache,
+                                                                         bool isYUpInFramebuffer)
+{
+    const auto &key = inCmd.m_shaderPathKey;
+    qCDebug(lcEffectSystem) << "    generating new shader pipeline for: " << key;
+
+    generator->beginProgram();
+
+    {
+        const QByteArray src = shaderLib->getShaderSource(inCmd.m_shaderPathKey, QSSGShaderCache::ShaderType::Vertex);
+        QSSGStageGeneratorBase *vStage = generator->getStage(QSSGShaderGeneratorStage::Vertex);
+        vStage->append(isYUpInFramebuffer ? effect_builtin_textureMapUV : effect_builtin_textureMapUVFlipped);
+        vStage->append(src);
+    }
+    {
+        const QByteArray src = shaderLib->getShaderSource(inCmd.m_shaderPathKey, QSSGShaderCache::ShaderType::Fragment);
+        QSSGStageGeneratorBase *fStage = generator->getStage(QSSGShaderGeneratorStage::Fragment);
+        fStage->append(src);
+    }
+
+    return generator->compileGeneratedRhiShader(key,
+                                                ShaderFeatureSetList(),
+                                                shaderLib,
+                                                shaderCache,
+                                                QSSGRhiShaderPipeline::UsedWithoutIa);
+}
+
 void QSSGRhiEffectSystem::bindShaderCmd(const QSSGBindShader *inCmd)
 {
     m_currentTextures.clear();
     m_pendingClears.clear();
 
-    // now we need a proper unique key (unique in the scene), the filenames are not sufficient
-    const QByteArray key = inCmd->m_shaderPathKey
-            + QByteArray::number(quintptr(inCmd->m_effect))
-            + QByteArray::number(inCmd->m_passIndex);
-
-    auto it = m_shaderPipelines.find(key);
-    if (it != m_shaderPipelines.end()) {
-        m_currentShaderPipeline = it.value().data();
-
-        const void *cacheKey1 = reinterpret_cast<const void *>(this);
-        const void *cacheKey2 = reinterpret_cast<const void *>(qintptr(m_currentUbufIndex));
-        QSSGRhiDrawCallData &dcd = m_rhiContext->drawCallData({ cacheKey1, cacheKey2, nullptr, 0, QSSGRhiDrawCallDataKey::Effects });
-        m_currentShaderPipeline->ensureCombinedMainLightsUniformBuffer(&dcd.ubuf);
-        m_currentUBufData = dcd.ubuf->beginFullDynamicBufferUpdateForCurrentFrame();
-
-        return;
-    }
-
-    qCDebug(lcEffectSystem) << "    generating new shader pipeline for lookup key" << key;
-
-    const QSSGRef<QSSGProgramGenerator> &generator = m_renderer->contextInterface()->shaderProgramGenerator();
-    generator->beginProgram();
-
-    QSSGShaderLibraryManager *shaderLib = m_renderer->contextInterface()->shaderLibraryManager().data();
-    {
-        const QByteArray src = shaderLib->getShaderSource(inCmd->m_shaderPathKey, QSSGShaderCache::ShaderType::Vertex);
-        QSSGStageGeneratorBase *vStage = generator->getStage(QSSGShaderGeneratorStage::Vertex);
-        QRhi *rhi = m_renderer->contextInterface()->rhiContext()->rhi();
-        if (rhi->isYUpInFramebuffer())
-            vStage->append(effect_builtin_textureMapUV);
-        else
-            vStage->append(effect_builtin_textureMapUVFlipped);
-        vStage->append(src);
-    }
-    {
-        const QByteArray src = shaderLib->getShaderSource(inCmd->m_shaderPathKey, QSSGShaderCache::ShaderType::Fragment);
-        QSSGStageGeneratorBase *fStage = generator->getStage(QSSGShaderGeneratorStage::Fragment);
-        fStage->append(src);
-    }
-
+    QRhi *rhi = m_renderer->contextInterface()->rhiContext()->rhi();
+    const auto &shaderLib = m_renderer->contextInterface()->shaderLibraryManager();
     const auto &shaderCache = m_renderer->contextInterface()->shaderCache();
-    QSSGRef<QSSGRhiShaderPipeline> stages = generator->compileGeneratedRhiShader(key, ShaderFeatureSetList(), shaderLib, shaderCache,
-                                                                                 QSSGRhiShaderPipeline::UsedWithoutIa);
-    if (stages) {
-        m_shaderPipelines.insert(key, stages);
-        m_currentShaderPipeline = m_shaderPipelines[key].data();
 
+    const QByteArray &key = inCmd->m_shaderPathKey;
+    const auto &hkey = inCmd->m_hkey;
+    Q_ASSERT(hkey != 0);
+    // now we need a proper "unique" key (unique in the scene), the filenames are not sufficient
+    const auto rkey = hkey ^ quintptr(inCmd) ^ m_currentUbufIndex;
+
+    // look for a runtime pipeline
+    const auto it = m_shaderPipelines.constFind(rkey);
+    if (it != m_shaderPipelines.cend())
+        m_currentShaderPipeline = (*it).data();
+
+    // Check if there's a build-time genereated entry for this effect
+    if (!m_currentShaderPipeline) {
+        const auto &shaderEntries = shaderLib->m_shaderEntries;
+        const auto foundIt = shaderEntries.constFind(QQsbCollection::Entry{hkey});
+        if (foundIt != shaderEntries.cend()) {
+            const auto &shader = shaderCache->loadGeneratedShader(key, *foundIt);
+            m_shaderPipelines.insert(rkey, shader);
+            m_currentShaderPipeline = shader.data();
+        }
+    }
+
+    // Final option, generate the shader pipeline
+    const QSSGRef<QSSGProgramGenerator> &generator = m_renderer->contextInterface()->shaderProgramGenerator();
+    if (auto stages = buildShaderForEffect(*inCmd, generator, shaderLib, shaderCache, rhi->isYUpInFramebuffer())) {
+        m_shaderPipelines.insert(rkey, stages);
+        m_currentShaderPipeline = stages.data();
+    }
+
+    if (m_currentShaderPipeline) {
         const void *cacheKey1 = reinterpret_cast<const void *>(this);
         const void *cacheKey2 = reinterpret_cast<const void *>(qintptr(m_currentUbufIndex));
         QSSGRhiDrawCallData &dcd = m_rhiContext->drawCallData({ cacheKey1, cacheKey2, nullptr, 0, QSSGRhiDrawCallDataKey::Effects });
@@ -489,7 +500,9 @@ void QSSGRhiEffectSystem::renderCmd(QSSGRhiEffectTexture *inTexture, QSSGRhiEffe
         // anyway when rendering the quad.
         if (rt != target->renderTarget) {
             cb->beginPass(rt, Qt::transparent, { 1.0f, 0 }, nullptr, QSSGRhiContext::commonPassFlags());
+            QSSGRHICTX_STAT(m_rhiContext, beginRenderPass(rt));
             cb->endPass();
+            QSSGRHICTX_STAT(m_rhiContext, endRenderPass());
         }
     }
     m_pendingClears.clear();
